@@ -10,9 +10,9 @@ import (
 	"time"
 
 	buildahcli "github.com/containers/buildah/pkg/cli"
-	"github.com/containers/buildah/pkg/formats"
 	"github.com/containers/buildah/pkg/parse"
 	"github.com/containers/common/libimage"
+	"github.com/containers/common/pkg/formats"
 	"github.com/docker/go-units"
 	"github.com/spf13/cobra"
 )
@@ -142,11 +142,29 @@ func imagesCmd(c *cobra.Command, args []string, iopts *imageResults) error {
 		options.Filters = append(options.Filters, "intermediate=false")
 	}
 
-	images, err := runtime.ListImages(ctx, args, options)
+	images, err := runtime.ListImages(ctx, options)
 	if err != nil {
 		return err
 	}
 
+	if len(args) > 0 {
+		imagesMatchName, err := runtime.ListImagesByNames(args)
+		if err != nil {
+			return err
+		}
+		imagesIDs := map[string]struct{}{}
+		for _, image := range imagesMatchName {
+			imagesIDs[image.ID()] = struct{}{}
+		}
+		var imagesMatchNameAndFilter []*libimage.Image
+		for _, image := range images {
+			if _, ok := imagesIDs[image.ID()]; ok {
+				imagesMatchNameAndFilter = append(imagesMatchNameAndFilter, image)
+			}
+		}
+		images = imagesMatchNameAndFilter
+	}
+
 	if iopts.quiet && iopts.format != "" {
 		return errors.New("quiet and format are mutually exclusive")
 	}
@@ -171,7 +189,7 @@ func imagesCmd(c *cobra.Command, args []string, iopts *imageResults) error {
 
 func outputHeader(opts imageOptions) string {
 	if opts.format != "" {
-		return strings.Replace(opts.format, `\t`, "\t", -1)
+		return strings.ReplaceAll(opts.format, `\t`, "\t")
 	}
 	if opts.quiet {
 		return formats.IDString
@@ -309,10 +327,10 @@ func truncateID(id string, truncate bool) string {
 	return id
 }
 
-func imagesToGeneric(templParams []imageOutputParams) (genericParams []interface{}) {
+func imagesToGeneric(templParams []imageOutputParams) (genericParams []any) {
 	if len(templParams) > 0 {
 		for _, v := range templParams {
-			genericParams = append(genericParams, interface{}(v))
+			genericParams = append(genericParams, any(v))
 		}
 	}
 	return genericParams
