@@ -2,6 +2,7 @@ package tmpdir
 
 import (
 	"os"
+	"path/filepath"
 	"testing"
 
 	"github.com/containers/common/pkg/config"
@@ -10,6 +11,7 @@ import (
 )
 
 func TestGetTempDir(t *testing.T) {
+	t.Parallel()
 	// test default
 	err := os.Unsetenv("TMPDIR")
 	require.NoError(t, err)
@@ -26,6 +28,14 @@ func TestGetTempDir(t *testing.T) {
 	err = os.Unsetenv("TMPDIR")
 	require.NoError(t, err)
 
+	// relative TMPDIR should be automatically converted to absolute
+	err = os.Setenv("TMPDIR", ".")
+	require.NoError(t, err)
+	tmpdir = GetTempDir()
+	assert.True(t, filepath.IsAbs(tmpdir), "path from GetTempDir should always be absolute")
+	err = os.Unsetenv("TMPDIR")
+	require.NoError(t, err)
+
 	f, err := os.CreateTemp("", "containers.conf-")
 	require.NoError(t, err)
 	// close and remove the temporary file at the end of the program
@@ -45,5 +55,4 @@ func TestGetTempDir(t *testing.T) {
 	require.NoError(t, err)
 	tmpdir = GetTempDir()
 	assert.Equal(t, "/mnt", tmpdir)
-
 }
