@@ -4,9 +4,12 @@ import (
 	"testing"
 
 	"github.com/containers/storage"
+	"github.com/openshift/imagebuilder"
+	"github.com/stretchr/testify/assert"
 )
 
 func TestGetImageName(t *testing.T) {
+	t.Parallel()
 	tt := []struct {
 		caseName string
 		name     string
@@ -26,3 +29,9 @@ func TestGetImageName(t *testing.T) {
 		}
 	}
 }
+
+func TestNoBaseImageSpecifierIsScratch(t *testing.T) {
+	t.Parallel()
+	assert.Equal(t, "scratch", imagebuilder.NoBaseImageSpecifier) // juuuuust in case
+	assert.Equal(t, "scratch", BaseImageFakeName)
+}
